----------------------------------------------------------------------------------------
--   Script Name:	Symmetry (Core Functions)
--	 Version:		v0.11
--	 Started:       02/18/2001
--	 Last Modified: 02/18/2001
--   Description:	Select Symmetrical Sub-Objects in PolyObjects.
--	 Documentation: Visit http://www.gfxcentral.com/bobo/mxs4/symmetry
--	 Code by:		Borislav Petrov [Boboland]		bobo@email.archlab.tuwien.ac.at
----------------------------------------------------------------------------------------

(
-----------------------
--USER CONTROL VALUES--
-----------------------
persistent global symmetry_divisions = 10			--edit this to change voxel count. Min 2, Max ca. 20

persistent global symmetry_threshold_value = 0.1	--edit this to change search threshold. 
													--May not be lower than the min distance between two verts in the mesh 
													--or sel. might be incomplete.

persistent global symmetry_poly_obj = undefined
persistent global symmetry_force_rebuild_flag = true
persistent global symmetry_search_mode = 1			--experimental, don't change
persistent global symmetry_display_info = false		--set this to toggle info in Listener.


-------------------------------
--INTERNAL PERSISTENT GLOBALS--
-------------------------------

persistent global symmetry_automode_x = false
persistent global symmetry_automode_y = false
persistent global symmetry_automode_z = false

persistent global symmetry_allowed = true

persistent global symmetry_min_x 
persistent global symmetry_min_y 
persistent global symmetry_min_z 
persistent global symmetry_max_x 
persistent global symmetry_max_y 
persistent global symmetry_max_z 

persistent global symmetry_cube_x_size 
persistent global symmetry_cube_y_size 
persistent global symmetry_cube_z_size 
persistent global symmetry_voxel_array_limits
persistent global symmetry_voxel_array_vertices
persistent global symmetry_interactive_vertex_array 
persistent global symmetry_current_active_axis
persistent global symmetry_current_class

------------------------
--FUNCTION DEFINITIONS--
------------------------
global symmetry_check_class 
global symmetry_build_symmetry_database
global symmetry_force_data_rebuild 
global symmetry_about_axis 
global symmetry_read_cube 
global symmetry_allowed_function


---------------------------------
--CHECK CLASS OF CURRENT OBJECT--
---------------------------------
fn symmetry_check_class =
(
	symmetry_current_class = 0

	if isValidNode symmetry_poly_obj then
	(
	if classof symmetry_poly_obj == Editable_mesh then
	(
	symmetry_current_class = 2
	if symmetry_display_info then format "Symmetry In EditMesh Mode\n"
	)
	if classof symmetry_poly_obj == Editable_Poly then
	(
	symmetry_current_class = 1
	if symmetry_display_info then format "Symmetry In PolyObject Mode\n"
	)
	)
	else
	(
	global symmetry_automode_x = false
	global symmetry_automode_y = false
	global symmetry_automode_z = false	
	)
)


fn symmetry_allowed_function = 
(
selection.count == 1 and (classof selection[1] == Editable_mesh or classof selection[1] == Editable_Poly )
)


------------------------------------
--BUILD SPATIAL PARTITION DATABASE--
------------------------------------
fn symmetry_build_database =
(
case symmetry_current_class of
(
1: num_verts = polyop.getnumverts symmetry_poly_obj
2: num_verts = getnumverts symmetry_poly_obj
)

symmetry_voxel_array_limits =#()
symmetry_voxel_array_vertices =#()

symmetry_min_x = 1000
symmetry_min_y = 1000
symmetry_min_z = 1000
symmetry_max_x = -1000
symmetry_max_y = -1000
symmetry_max_z = -1000

for i = 1 to num_verts do
(
case symmetry_current_class of
(
1: vert = in coordsys local polyop.getvert symmetry_poly_obj i
2: vert = in coordsys local getvert symmetry_poly_obj i
)

if vert.x < symmetry_min_x then symmetry_min_x = vert.x
if vert.y < symmetry_min_y then symmetry_min_y = vert.y
if vert.z < symmetry_min_z then symmetry_min_z = vert.z
if vert.x > symmetry_max_x then symmetry_max_x = vert.x
if vert.y > symmetry_max_y then symmetry_max_y = vert.y
if vert.z > symmetry_max_z then symmetry_max_z = vert.z
)--end i

symmetry_min_x -= 1
symmetry_min_y -= 1
symmetry_min_z -= 1
symmetry_max_x += 1
symmetry_max_y += 1
symmetry_max_z += 1

symmetry_x_size = symmetry_max_x-symmetry_min_x
symmetry_y_size = symmetry_max_y-symmetry_min_y
symmetry_z_size = symmetry_max_z-symmetry_min_z

symmetry_cube_x_size = symmetry_x_size / symmetry_divisions
symmetry_cube_y_size = symmetry_y_size / symmetry_divisions
symmetry_cube_z_size = symmetry_z_size / symmetry_divisions

for x = 1 to symmetry_divisions do
(
symmetry_voxel_array_limits[x]=#()
symmetry_voxel_array_vertices[x]=#()
for y = 1 to symmetry_divisions do
(
symmetry_voxel_array_limits[x][y]=#()
symmetry_voxel_array_vertices[x][y]=#()
for z = 1 to symmetry_divisions do
(
min_limit = [(symmetry_min_x+(x-1)*symmetry_cube_x_size), (symmetry_min_y+(y-1)*symmetry_cube_y_size), (symmetry_min_z+(z-1)*symmetry_cube_z_size)]
max_limit = [(symmetry_min_x+(x)*symmetry_cube_x_size), (symmetry_min_y+(y)*symmetry_cube_y_size), (symmetry_min_z+(z)*symmetry_cube_z_size)]
symmetry_voxel_array_limits[x][y][z]=#(min_limit,max_limit)
symmetry_voxel_array_vertices[x][y][z]=#()
)--end z
)--end y
)--end x

for i = 1 to num_verts do
(
case symmetry_current_class of
(
1: vert = in coordsys local polyop.getvert symmetry_poly_obj i
2: vert = in coordsys local getvert symmetry_poly_obj i
)
div_x = ceil ((vert.x - symmetry_min_x)/ symmetry_cube_x_size )
div_y = ceil ((vert.y - symmetry_min_y)/ symmetry_cube_y_size )
div_z = ceil ((vert.z - symmetry_min_z)/ symmetry_cube_z_size )
if div_x == 0 then div_x = 1
if div_y == 0 then div_y = 1
if div_z == 0 then div_z = 1
append symmetry_voxel_array_vertices[div_x][div_y][div_z] #(i, vert)
)

symmetry_force_rebuild_flag = false
)--end fn




-----------------------------------
--FIND VERTEX INSIDE A  SUB-SPACE--
-----------------------------------
fn symmetry_read_cube vert div_x div_y div_z mode thresh=
(
try
(
return_value = 0
min_dist = 1000000.0
cube_array = symmetry_voxel_array_vertices[div_x][div_y][div_z]
for v in cube_array do 
(
if mode == 1 and abs(v[2].x - vert.x) < thresh and abs(v[2].y - vert.y) < thresh and abs(v[2].z - vert.z) < thresh then
	(
	return v[1]
	)


if mode == 2 and abs(v[2].x - vert.x) < thresh and abs(v[2].y - vert.y) < thresh and abs(v[2].z - vert.z) < thresh then
	(
	if length(v[2]-vert) < min_dist then 
		(
		min_dist = length(v[2]-vert)
		return_value = v[1]
		)
	)
	
if mode == 3 then
	(
	if length(v[2]-vert) < min_dist then 
		(
		min_dist = length(v[2]-vert)
		return_value = v[1]
		)
	)
	
)--end v
)
catch()
return_value
)

----------------------------------------
--FIND SUB-SPACE BY SPATIAL COORDINATE--
----------------------------------------
fn symmetry_find_vertex vert axis mode thresh=
(
case axis of
(
1: vert = [-vert.x, vert.y, vert.z]
2: vert = [vert.x, -vert.y, vert.z]
3: vert = [vert.x, vert.y, -vert.z]
)
div_x = ceil ((vert.x - symmetry_min_x)/ symmetry_cube_x_size )
div_y = ceil ((vert.y - symmetry_min_y)/ symmetry_cube_y_size )
div_z = ceil ((vert.z - symmetry_min_z)/ symmetry_cube_z_size )
if div_x == 0 then div_x = 1
if div_y == 0 then div_y = 1
if div_z == 0 then div_z = 1
result = symmetry_read_cube vert div_x div_y div_z mode thresh
result
)--end fn


----------------------------------------
--FORCE A DATA REBUILD, REPORT RESULTS--
----------------------------------------
fn symmetry_force_data_rebuild =
(
	if selection.count == 1 then
	(
	global symmetry_poly_obj = selection[1] 
	symmetry_check_class()
	s_t = timestamp()
	symmetry_build_database()
	e_t = timestamp()
	if symmetry_display_info then 
	(
	format "\nDatabase Rebuilt in: %ms.\n" (e_t-s_t)
	format "Sub-Space Divisions: % \n" (symmetry_divisions)
	format "Sub-Space Count: % \n\n" (symmetry_divisions^3)
	)
	)
	else
	(
	symmetry_allowed = true
	)
)

----------------------------
--SYMMETRY SELECT FUNCTION--
----------------------------
fn symmetry_about_axis axis do_select mode thresh =
(
if symmetry_allowed then
(
setWaitCursor()
if not isValidNode symmetry_poly_obj then symmetry_force_rebuild_flag = true
if selection[1] != symmetry_poly_obj then symmetry_force_rebuild_flag = true
if symmetry_force_rebuild_flag then symmetry_force_data_rebuild()

check_count = 0
sym_point_found = 0
s_t = timestamp()

case SubObjectLevel of
(
1: (
	case symmetry_current_class of
	(
	1: (
		vert_sel = polyop.getvertselection symmetry_poly_obj
		base_sel = polyop.getvertselection symmetry_poly_obj
		)
	2: (
		vert_sel = getvertselection symmetry_poly_obj
		base_sel = getvertselection symmetry_poly_obj
		)
	)--end case

	for v in base_sel do
	(
	check_count +=1
	case symmetry_current_class of
	(
	1: result = symmetry_find_vertex (in coordsys local polyop.getvert symmetry_poly_obj v) axis mode thresh
	2: result = symmetry_find_vertex (in coordsys local getvert symmetry_poly_obj v) axis mode thresh
	)	
	if result > 0 then 
	(
	append vert_sel result
	if base_sel[result] == true then base_sel[result] = false
	sym_point_found += 1
	)
	)--end v loop	
	
	)--end vertex case
2: (
	case symmetry_current_class of
	(
	1: (
		vert_sel = #{}
		base_sel = #{}
		edge_sel = polyop.getedgeselection symmetry_poly_obj
		edge_base_sel = polyop.getedgeselection symmetry_poly_obj
		for e in edge_sel do 
			(
			edge_verts = (polyop.getEdgeVerts symmetry_poly_obj e)
					result1 = symmetry_find_vertex (in coordsys local polyop.getvert symmetry_poly_obj edge_verts[1]) axis mode thresh
					result2 = symmetry_find_vertex (in coordsys local polyop.getvert symmetry_poly_obj edge_verts[2]) axis mode thresh
					check_count += 2
					result_array = #(result1,result2)
				
					if result1 > 0 and result2 > 0 then 
						(
						sym_point_found += 2
						potential_edges = polyop.getEdgesUsingVert symmetry_poly_obj #{result1,result2}
						for pe in potential_edges do
							(
							used_verts = (polyop.getEdgeVerts symmetry_poly_obj pe)
							if findItem result_array used_verts[1] > 0 and findItem result_array used_verts[2] > 0 then
								(
								append edge_base_sel pe
								)
							)--end pe loop
						)--end if	
			)--end e loop
		)--end 1	
	)--end case
	)--end edge case

4: (
	case symmetry_current_class of
	(
	1: (
		vert_sel = #{}
		base_sel = #{}
		poly_sel = polyop.getFaceSelection symmetry_poly_obj
		poly_base_sel = polyop.getFaceSelection symmetry_poly_obj
		for f in poly_sel do 
		(
			poly_verts = (polyop.getFaceVerts symmetry_poly_obj f)
			result_array = #()
			result_bitarray = #{}
			for v in poly_verts do
			(
			result = symmetry_find_vertex (in coordsys local polyop.getvert symmetry_poly_obj v) axis mode thresh
			check_count += 1
			if result > 0 then 
			(
				sym_point_found += 1
				append result_array result 
				append result_bitarray result
			)
			else 
			(
				result_array = #()
			)	
			)--end v loop
				
			if result_array.count == poly_verts.count then 
				(
				potential_faces = polyop.getFacesUsingVert symmetry_poly_obj result_bitarray
				for pf in potential_faces do
					(
						used_verts = (polyop.getFaceVerts symmetry_poly_obj pf)
						test_count = 0
						for uf in used_verts do
						(
							if findItem result_array uf > 0 then
							(
							test_count +=1
							)
						)--end uf loop
						if test_count == used_verts.count then append poly_base_sel pf
					)--end pf loop
				)--end if result	
			)--end f loop
		)--end 1
	)--end case current_class
	)--end face case
	
)--end case


e_t = timestamp()

if symmetry_display_info then
(
format "% Queries.\n" check_count
format "% Symmetry Points Found.\n" sym_point_found
format "Search Time: %ms.\n" (e_t-s_t)
)

case subObjectLevel of
(
1: (
	case symmetry_current_class of
	(
	1: (
		select symmetry_poly_obj.verts[vert_sel]
		)
	2: setvertselection symmetry_poly_obj vert_sel
	)
	)
2: (
	case symmetry_current_class of
	(
	1: (
		select symmetry_poly_obj.edges[edge_base_sel]
		)
	2: ()
	)	
	)
4: (
	case symmetry_current_class of
	(
	1: (
		select symmetry_poly_obj.faces[poly_base_sel]
		)
	)--end case	
	)--end 4
	
)--end case
)
setArrowCursor()
)--end fn


)--end script
